// Proxy.cpp
#include "Proxy.h"
#include <iostream>
#include <thread>
#include <sstream>
#include <fmt/format.h>
#include <boost/asio/ssl.hpp>
#include <boost/asio/ip/tcp.hpp>
#include "Utils.h"
#include "Configfile.h"

extern Configfile cfg;

namespace {
std::string resolveHostname(const std::string& ip_address) {
    try {
        boost::asio::io_context io_context;
        boost::asio::ip::tcp::resolver resolver(io_context);
        boost::asio::ip::address addr = boost::asio::ip::make_address(ip_address);
        boost::asio::ip::tcp::endpoint ep(addr, 0);
        auto results = resolver.resolve(ep.address().to_string(), "");
        if (results.begin() != results.end()) {
            return results.begin()->host_name();
        }
    } catch (const std::exception&) {}
    return "";
}
}

Proxy::Proxy() 
    : io_context_(),
      ssl_context(boost::asio::ssl::context::tlsv12),
      outside_socket_(nullptr) {
}

void Proxy::setOutside(Socket& socket) {
    outside_socket_ = &socket;
}

void Proxy::run(const std::string& peer_address) {
    if (!outside_socket_) {
        throw std::runtime_error("Outside socket not set");
    }

    boost::asio::ssl::stream<boost::asio::ip::tcp::socket> inside(io_context_, ssl_context);
    // Original comments and variables retained
    const std::string empty_str;
    std::string from = empty_str;
    std::string to = empty_str;
    std::string ehlostr = empty_str;
    std::string resolvedname = empty_str;
    unsigned char last_state = SMTP_STATE_WAIT_FOR_HELO;
    long unimplemented_requests = 0;

    #ifdef HAVE_SPF
    SpfChecker spf_checker;
    #endif

    try {
        // Resolve hostname using the Boost resolver
        resolvedname = resolveHostname(peer_address);

        // Configure SSL contexts
        boost::asio::ssl::context ssl_context(boost::asio::ssl::context::tlsv12);
        ssl_context.set_verify_mode(boost::asio::ssl::verify_none);

        bool throttled = cfg.getThrottle(); // Start with a throttled connection
        bool authenticated = false; // Start with a non-authenticated connection
        bool esmtp = false;
        std::string strtemp;
        std::string hermes_status = "unknown";

        // Check whitelist
        if (!cfg.getDnsWhitelistDomains().empty() &&
            Utils::listed_on_dns_lists(cfg.getDnsWhitelistDomains(), cfg.getDnsWhitelistPercentage(), peer_address)) {
            authenticated = true;
            hermes_status = "whitelisted";
            if (cfg.getWhitelistedDisablesEverything()) {
                throttled = false;
            }
        }

        std::string peer_addr_copy = peer_address;
        if (cfg.getWhitelistedDisablesEverything() && Utils::whitelisted(cfg.getDatabaseFile(), peer_addr_copy)) {
            throttled = false;
            authenticated = true;
        } else {
            if (!cfg.getAllowDataBeforeBanner()) {
                std::this_thread::sleep_for(std::chrono::seconds(cfg.getBannerDelayTime()));

                // Check if data is waiting before server banner
                if (outside_socket_->canRead(0.0)) {
                    std::cout << fmt::format("421 (data_before_banner) (ip:{})\n", peer_address);
                    std::this_thread::sleep_for(std::chrono::seconds(20));

                    // Write rejection message
                    std::string rejection_msg = fmt::format("421 Stop sending data before we show you the banner\r\n");
                    outside_socket_->writeBytes(const_cast<char*>(rejection_msg.c_str()), rejection_msg.length());
                    return;
                }
            }
        }

        // Connect to the inside server
        boost::asio::ip::tcp::resolver resolver(io_context_);
        boost::asio::ip::tcp::resolver::results_type endpoints = 
            resolver.resolve(cfg.getServerHost(), std::to_string(cfg.getServerPort()));

        boost::asio::connect(inside.lowest_layer(), endpoints);

        // SSL setup
        if (cfg.getOutgoingSsl()) {
            inside.set_verify_mode(boost::asio::ssl::verify_none);
            inside.handshake(boost::asio::ssl::stream_base::client);
        }

        if (cfg.getIncomingSsl()) {
            #ifdef HAVE_SSL
            outside_socket_->prepareSSL(true);
            outside_socket_->startSSL(true);
            #endif
        }

        // Communication buffers
        char read_buffer[4096];
        ssize_t bytes_read;

        // Main loop for communication
        while (!outside_socket_->isClosed() && !inside.lowest_layer().is_open()) {
            // Check if the client wants to send something to the server
            if (outside_socket_->canRead(1.0)) {
                bytes_read = outside_socket_->readBytes(read_buffer, sizeof(read_buffer));
                if (bytes_read > 0) {
                    strtemp = std::string(read_buffer, bytes_read);

                    std::string cmd_lower = Utils::strtolower(strtemp.substr(0, std::min<size_t>(10, strtemp.length())));
                    if (strtemp.length() > 10 && "mail from:" == cmd_lower) {
                        from = std::string(Utils::getmail(strtemp));
                        last_state = SMTP_STATE_WAIT_FOR_RCPTTO;
                    }

                    cmd_lower = Utils::strtolower(strtemp.substr(0, std::min<size_t>(4, strtemp.length())));
                    if ("ehlo" == cmd_lower) esmtp = true;

                    if (strtemp.length() > 4 && ("ehlo" == cmd_lower ||
                            "helo" == cmd_lower)) {
                        ehlostr = std::string(Utils::trim(strtemp.substr(5)));
                        last_state = SMTP_STATE_WAIT_FOR_MAILFROM;
                    }

                    // RCPT TO handling with comprehensive checks
                    cmd_lower = Utils::strtolower(strtemp.substr(0, std::min<size_t>(8, strtemp.length())));
                    if (strtemp.length() > 8 && "rcpt to:" == cmd_lower) {
                        std::string mechanism;
                        std::string message;
                        to = std::string(Utils::getmail(strtemp));

                        // Construct log string
                        std::string strlog = fmt::format("from {} (ip:{}, hostname:{}, {} {}) -> to {}", 
                            from, peer_address, resolvedname, (esmtp ? "ehlo" : "helo"), ehlostr, to);

                        // Greylisting check
                        std::string code = "250";
                        std::string peer_addr_copy = peer_address;
                        std::string from_copy = from;
                        std::string to_copy = to;
                        if (cfg.getGreylist() && !authenticated && 
                            Utils::greylist(cfg.getDatabaseFile(), peer_addr_copy, from_copy, to_copy)) {
                            code = "421";
                            mechanism = "greylist";
                            message = fmt::format("{} Greylisted!! Please try again in a few minutes.", code);
                            std::cout << fmt::format("checking {}\n", mechanism);
                        }
                        // SPF Check
                        #ifdef HAVE_SPF
                        else if (cfg.getQuerySpf() && !authenticated && 
                                 !spf_checker.query(peer_address, ehlostr, from)) {
                            code = cfg.getAddStatusHeader() ? "250" : 
                                (cfg.getReturnTempErrorOnReject() ? "421" : "550");
                            mechanism = "spf";
                            message = fmt::format("{} You do not seem to be allowed to send email for that particular domain.", code);
                            std::cout << fmt::format("checking {}\n", mechanism);
                        }
                        #endif
                        // Blacklist check
                        else if (!authenticated && 
                                 Utils::blacklisted(cfg.getDatabaseFile(), peer_addr_copy, to_copy)) {
                            code = cfg.getReturnTempErrorOnReject() ? "421" : "550";
                            mechanism = "allowed-domain-per-ip";
                            message = fmt::format("{} You do not seem to be allowed to send email to that particular domain from that address.", code);
                            std::cout << fmt::format("checking {}\n", mechanism);
                        }
                        // DNS Blacklist check
                        else if (!cfg.getDnsBlacklistDomains().empty() && !authenticated &&
                                 Utils::listed_on_dns_lists(cfg.getDnsBlacklistDomains(), 
                                                            cfg.getDnsBlacklistPercentage(), 
                                                            peer_address)) {
                            code = cfg.getAddStatusHeader() ? "250" : 
                                (cfg.getReturnTempErrorOnReject() ? "421" : "550");
                            mechanism = "dnsbl";
                            message = fmt::format("{} You are listed on some DNS blacklists. Get delisted before trying to send us email.", code);
                            std::cout << fmt::format("checking {}\n", mechanism);
                        }
                        // Reverse DNS check
                        else if (cfg.getRejectNoReverseResolution() && !authenticated && resolvedname.empty()) {
                            code = cfg.getReturnTempErrorOnReject() ? "421" : "550";
                            mechanism = "no reverse resolution";
                            message = fmt::format("{} Your IP address does not resolve to a hostname.", code);
                            std::cout << fmt::format("checking {}\n", mechanism);
                        }
                        // HELO/Reverse name check
                        else if (cfg.getCheckHeloAgainstReverse() && !authenticated && ehlostr != resolvedname) {
                            code = cfg.getReturnTempErrorOnReject() ? "421" : "550";
                            mechanism = "helo differs from resolved name";
                            message = fmt::format("{} Your IP hostname doesn't match your envelope hostname.", code);
                            std::cout << fmt::format("checking {}\n", mechanism);
                        }

                        // Prepare log message
                        std::string log_str = strlog;
                        if (!mechanism.empty()) {
                            log_str = fmt::format("({}) {}", mechanism, log_str);
                        }
                        std::cout << fmt::format("{} {}\n", code, log_str);

                        // Handle rejection
                        if (code != "250") {
                            // Close inside connection
                            inside.lowest_layer().close();
                            
                            // Delay to annoy spammers
                            std::this_thread::sleep_for(std::chrono::seconds(20));
                            
                            // Send rejection message
                            std::string rejection_msg = fmt::format("{}\r\n", message);
                            outside_socket_->writeBytes(const_cast<char*>(rejection_msg.c_str()), rejection_msg.length());
                            return;
                        }

                        last_state = SMTP_STATE_WAIT_FOR_DATA;
                    }

                    // Send to inside server
                    boost::asio::async_write(inside, boost::asio::buffer(strtemp),
                        [](const boost::system::error_code& /*error*/, std::size_t /*bytes_transferred*/) {});
                }
            }

            // Check if the server wants to send something to the client
            boost::system::error_code ec;
            size_t server_available = inside.lowest_layer().available(ec);
            if (!ec && server_available > 0) {
                std::vector<char> server_buffer(server_available);
                size_t bytes_read = boost::asio::read(inside, boost::asio::buffer(server_buffer), ec);
                if (!ec && bytes_read > 0) {
                    outside_socket_->writeBytes(server_buffer.data(), bytes_read);
                }
            }

            // Run io_context to process async operations
            io_context_.poll();

            // Throttling
            if (throttled) {
                std::this_thread::sleep_for(std::chrono::seconds(cfg.getThrottlingTime()));
            }
        }
    }
    catch (const boost::system::system_error& e) {
        std::cerr << fmt::format("Boost.Asio error: {}\n", e.what());
    }
    catch (const std::exception& e) {
        std::cerr << fmt::format("Standard exception: {}\n", e.what());
    }
}
